#include "itkImageFileReader.h"
#include "itkErodeMaskImageFilter.h"
#include "itkErodeMaskImageFilter2.h"
#include "itkImageFileWriter.h"
#include "itkOptGrayscaleErodeImageFilter.h"
#include "itkFlatStructuringElement.h"
#include "itkBinaryErodeImageFilter.h"
#include "itkBinaryBallStructuringElement.h"
#include "itkNeighborhood.h"

#include "itkIterativeImageFilter.h"
#include "itkCastImageFilter.h"
#include "itkBinaryContourImageFilter.h"
#include "itkSubtractImageFilter.h"

#include "itkParabolicErodeImageFilter.h"

#include <string>
#include <iostream>
#include <iomanip>
#include <ctime>

//-------------------------------------------------------------------------------------

int main( int argc, char *argv[] )
{
  typedef unsigned char PixelType;
  const unsigned int Dimension = 3;
  typedef itk::Image<PixelType,Dimension> ImageType;
  typedef itk::ImageFileReader<ImageType> ReaderType;
  typedef itk::ImageFileWriter<ImageType> WriterType;
  typedef itk::ErodeMaskImageFilter<ImageType> ErosionTypeOld;
  typedef itk::ErodeMaskImageFilter2<ImageType> ErosionTypeNew2;
  typedef ErosionTypeOld::ScheduleType   ScheduleType;
  typedef itk::FlatStructuringElement<Dimension>    KernelType;
  typedef itk::GrayscaleErodeImageFilter< ImageType,ImageType,KernelType > ErosionTypeNew;
  typedef ErosionTypeNew::RadiusType RadiusType;

  //typedef itk::BinaryBallStructuringElement<PixelType,Dimension> KernelType2;
  typedef itk::Neighborhood<PixelType,Dimension>  KernelType2;
  typedef itk::BinaryErodeImageFilter<ImageType,ImageType,KernelType2> ErosionTypeBinary;
  typedef KernelType2::RadiusType     RadiusType2;

  typedef itk::CastImageFilter< ImageType, ImageType > CastType;
  typedef itk::BinaryContourImageFilter< ImageType, ImageType > ContourType;
  typedef itk::SubtractImageFilter< ImageType, ImageType, ImageType > SubtractType;
  typedef itk::IterativeImageFilter< ImageType, ImageType > IterativeType;

  typedef itk::ParabolicErodeImageFilter< ImageType, ImageType > ParabolicType;

  /** Read */
  ReaderType::Pointer reader = ReaderType::New();
  reader->SetFileName( argv[1] );
  reader->Update();

  std::clock_t start;
  std::ostringstream makeFileName( "" );

  /** Write. */
  WriterType::Pointer writer = WriterType::New();
  writer->SetUseCompression( true );

  std::vector< unsigned int > radii( 6 );
  radii[ 0 ] = 1;
  radii[ 1 ] = 2;
  radii[ 2 ] = 4;
  radii[ 3 ] = 8;
  radii[ 4 ] = 16;
  radii[ 5 ] = 32;
//   std::vector< unsigned int > radii( 1 );
//   radii[ 0 ] = 32;
  for ( unsigned int i = 0; i < radii.size(); ++i )
  {
    
  /** Setup old erosion filter. */
  ScheduleType schedule( 1, Dimension );
  schedule.Fill( radii[ i ] - 1 );

  ErosionTypeOld::Pointer erosion_old = ErosionTypeOld::New();
  erosion_old->SetSchedule( schedule );
  erosion_old->SetResolutionLevel( 0 );
  erosion_old->SetInput( reader->GetOutput() );

  /** Setup the fastest one. */
  ErosionTypeNew2::Pointer erosion_new2 = ErosionTypeNew2::New();
  erosion_new2->SetSchedule( schedule );
  erosion_new2->SetResolutionLevel( 0 );
  erosion_new2->SetInput( reader->GetOutput() );

  /** Setup binary erosion filter. */
  RadiusType2 radius2;
  radius2.Fill( radii[ i ] );
  KernelType2 kernel2;
  //kernel2.SetRadius( radius2 );
  //kernel2.CreateStructuringElement();

  kernel2.SetRadius( radii[ i ] );
  for ( unsigned int j = 0; j < kernel2.Size(); ++j ) kernel2[ j ] = 1;

  ErosionTypeBinary::Pointer erosion_binary = ErosionTypeBinary::New();
  erosion_binary->SetForegroundValue( 1 );
  erosion_binary->SetBackgroundValue( 0 );
  erosion_binary->SetKernel( kernel2 );
  erosion_binary->SetInput( reader->GetOutput() );
  
  /** Setup new erosion filter. */
  ErosionTypeNew::Pointer erosion_new = ErosionTypeNew::New();
  RadiusType radius; radius.Fill( radii[ i ] );
  //erosion_new->SetRadius( radii[ i ] );
  erosion_new->SetKernel( itk::FlatStructuringElement<Dimension>::Box( radius ) );
  std::string whatkind = "Box";
  //erosion_new->SetKernel( itk::FlatStructuringElement<Dimension>::Ball( radius ) );
  //std::string whatkind = "Ball";
  erosion_new->SetInput( reader->GetOutput() );

  std::cout << whatkind << " radius: " << radii[ i ] << std::endl;

  /** Run old. *
  start = clock();
  erosion_old->Update();
  std::cerr << "  Elapsed time erosion_old: " << static_cast<double>(
    clock() - start ) / CLOCKS_PER_SEC << "s." << std::endl;
  writer->SetFileName( "output_old.mhd" );
  writer->SetInput( erosion_old->GetOutput() );
  writer->Update();

  /** Run binary. *
  start = clock();
  erosion_binary->Update();
  std::cerr << "  Elapsed time erosion_binary (single-threaded): " << static_cast<double>(
    clock() - start ) / CLOCKS_PER_SEC << "s." << std::endl;
  makeFileName.str( "" );
  makeFileName << "output_binary_r" << radii[ i ] << ".mhd";
  writer->SetFileName( makeFileName.str().c_str() );
  writer->SetInput( erosion_binary->GetOutput() );
  writer->Update();

  /** Run new == binary, but in masks language. *
  start = clock();
  erosion_new2->Update();
  std::cerr << "  Elapsed time erosion new2: " << static_cast<double>(
    clock() - start ) / CLOCKS_PER_SEC << "s." << std::endl;
  makeFileName.str( "" );
  makeFileName << "output_new2_r" << radii[ i ] << ".mhd";
  writer->SetFileName( makeFileName.str().c_str() );
  writer->SetInput( erosion_new2->GetOutput() );
  writer->Update();

  /** Run basic. *
  erosion_new->SetAlgorithm( ErosionTypeNew::BASIC );
  start = clock();
  erosion_new->Update();
  std::cerr << "  Elapsed time erosion basic: " << static_cast<double>(
    clock() - start ) / CLOCKS_PER_SEC << "s." << std::endl;
  makeFileName.str( "" );
  makeFileName << "output_basic_r" << radii[ i ] << ".mhd";
  writer->SetFileName( makeFileName.str().c_str() );
  writer->SetInput( erosion_new->GetOutput() );
  writer->Update();

  /** Run hist. *
  erosion_new->SetAlgorithm( ErosionTypeNew::HISTO );
  start = clock();
  erosion_new->Update();
  std::cerr << "  Elapsed time erosion HISTO: " << static_cast<double>(
    clock() - start ) / CLOCKS_PER_SEC << "s." << std::endl;
  makeFileName.str( "" );
  makeFileName << "output_histo_r" << radii[ i ] << ".mhd";
  writer->SetFileName( makeFileName.str().c_str() );
  writer->SetInput( erosion_new->GetOutput() );
  writer->Update();

  /** Run van herk et al. */
  bool couldRun = true;
  try{ erosion_new->SetAlgorithm( ErosionTypeNew::VHGW ); }
  catch ( itk::ExceptionObject & )
  {
    //std::cerr << "ERROR: " << excp << std::endl;
    //return 1;
    couldRun = false;
  }
  //erosion_new->Print( std::cerr );
  //erosion_new->GetKernel().Print( std::cerr );
  if ( couldRun )
  {
  start = clock();
  try{ erosion_new->Update(); }
  catch ( itk::ExceptionObject & excp )
  {
    std::cerr << "ERROR: " << excp << std::endl;
    return 1;
  }
  std::cerr << "  Elapsed time erosion VHGW (multi-threaded): " << static_cast<double>(
    clock() - start ) / CLOCKS_PER_SEC << "s." << std::endl;
  makeFileName.str( "" );
  makeFileName << "output_VHGW_r" << radii[ i ] << ".mhd";
  writer->SetFileName( makeFileName.str().c_str() );
  writer->SetInput( erosion_new->GetOutput() );
  writer->Update();
  }
  else
  {
    std::cerr << "  Elapsed time erosion VHGW: could not run" << std::endl;
  }

  /** Run Anchor. *
  couldRun = true;
  try{ erosion_new->SetAlgorithm( ErosionTypeNew::ANCHOR ); }
  catch ( itk::ExceptionObject & ){ couldRun = false; }
  if ( couldRun )
  {
    start = clock();
    erosion_new->Update();
    std::cerr << "  Elapsed time erosion Anchor: " << static_cast<double>(
      clock() - start ) / CLOCKS_PER_SEC << "s." << std::endl;
    makeFileName.str( "" );
    makeFileName << "output_ANCHOR_r" << radii[ i ] << ".mhd";
    writer->SetFileName( makeFileName.str().c_str() );
    writer->SetInput( erosion_new->GetOutput() );
    writer->Update();
  }
  else
  {
    std::cerr << "  Elapsed time erosion ANCHOR: could not run" << std::endl;
  }

  /** Erosion based on outer contour detection (binary only, multi-threaded). */
  CastType::Pointer cast = CastType::New();
  cast->SetInPlace( true );

  ContourType::Pointer contour = ContourType::New();
  contour->SetInput( cast->GetOutput() );
  contour->SetFullyConnected( true );
  contour->SetBackgroundValue( 0 );
  contour->SetForegroundValue( 1 );

  SubtractType::Pointer subtract = SubtractType::New();
  subtract->SetInput( 0, cast->GetOutput() );
  subtract->SetInput( 1, contour->GetOutput() );

  IterativeType::Pointer iterative = IterativeType::New();
  iterative->SetInput( reader->GetOutput() );
  iterative->SetInputFilter( cast );
  iterative->SetOutputFilter( subtract );
  iterative->SetNumberOfIterations( radii[ i ] );

  start = clock();
  iterative->Update();
  std::cerr << "  Elapsed time contour iterative (multi-threaded): " << static_cast<double>(
    clock() - start ) / CLOCKS_PER_SEC << "s." << std::endl;
  makeFileName.str( "" );
  makeFileName << "output_contour_r" << radii[ i ] << ".mhd";
  writer->SetFileName( makeFileName.str().c_str() );
  writer->SetInput( iterative->GetOutput() );
  writer->Update();

  /** Parabolic erode. */
  ParabolicType::Pointer parabolic = ParabolicType::New();
  parabolic->SetInput( reader->GetOutput() );
  //parabolic->SetUseImageSpacing( true );
  parabolic->SetScale( radii[ i ]*radii[ i ]/2 + 1 );
  start = clock();
  parabolic->Update();
  std::cerr << "  Elapsed time parabolic (single-threaded): " << static_cast<double>(
    clock() - start ) / CLOCKS_PER_SEC << "s." << std::endl;
  makeFileName.str( "" );
  makeFileName << "output_parabolic_r" << radii[ i ] << ".mhd";
  writer->SetFileName( makeFileName.str().c_str() );
  writer->SetInput( parabolic->GetOutput() );
  writer->Update();

  /***/
}
  /** Return a value. */
	return 0;

} // end main


